# jack-alsa-ctl
# Copyright (C) 2022 DCsunset
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from setuptools import setup
from pathlib import Path

repo_dir = Path(__file__).parent.absolute()

# get version
main_ns = {}
with open(repo_dir.joinpath("jack_alsa_ctl", "_version.py")) as f:
	exec(f.read(), main_ns)

# Long description
readme = repo_dir.joinpath('README.md')
with open(readme, "r") as f:
	long_description = f.read()

setup(
	name="jack-alsa-ctl",
	version=main_ns["__version__"],
	description="Control JACK audio with ALSA driver easily",
	long_description=long_description,
	long_description_content_type='text/markdown',
	author="DCsunset",
	author_email='DCsunset@protonmail.com',
	license="AGPL-3.0",
	url="https://github.com/DCsunset/jack-alsa-ctl",
	packages=["jack_alsa_ctl"],
	scripts=["bin/jack-alsa-ctl"],
	classifiers=[
		"Environment :: Console",
		"Topic :: Multimedia :: Sound/Audio",
		'Intended Audience :: End Users/Desktop',
		'Programming Language :: Python',
		"License :: OSI Approved :: GNU Affero General Public License v3"
	]
)



# jack-alsa-ctl
# Copyright (C) 2022 DCsunset
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import subprocess
import sys
import re

# List all sound devices
# max_num <= 0 means no limit
def list_devices(max_num: int = 0) -> list[str]:
	with open("/proc/asound/cards", "r") as f:
		out = f.read()
	regex = re.compile(r"\[\s*(.+?)\s*\]")
	cards = regex.findall(out)
	if max_num > 0:
		n = min(max_num, len(cards))
		# prioritize cards from the end
		# because they are external
		cards = cards[-n:]
	return cards
   

# Get current jack2 device using jack_control
def get_jack_device():
	try:
		res = subprocess.run(["jack_control", "dp"], capture_output=True)
		if res.returncode != 0:
			return "0"
		lines = res.stdout.decode("utf-8").splitlines()
		for line in lines:
			l = line.strip()
			if l.startswith("device:"):
				# l[:-1] removes the last parathesis
				return l[:-1].split(":")[-1]
	except Exception as e:
		print(e, file=sys.stderr)
	# return default device on failure
	return "0"

# Set/change current jack device
def set_jack_device_cmd(dev: str) -> list[str]:
	return [
		f"jack_control dps device hw:{dev}",
		# Restart jack after setting dev
		"jack_control stop",
		"jack_control start"
	]

# Get scontrols of a sound card
def get_amixer_scontrols(card: str) -> list[str]:
	try:
		res = subprocess.run(["amixer", "-c", card, "scontrols"], capture_output=True)
		if res.returncode != 0:
			return []
			
		names = []
		lines = res.stdout.decode("utf-8").splitlines()
		p = re.compile(r"'(.+)'")
		for l in lines:
			# find the last occurrence of the quoted name
			name = p.search(l).groups()[-1]
			names.append(name)
		return names
	except Exception as e:
		print(e, file=sys.stderr)

	# return empty on failure
	return []

	
# Get current volume scontrol and card for amixer
def get_current_control() -> tuple[str, str]:
	card = get_jack_device()
	scontrols = get_amixer_scontrols(card)

	# possible scontrols for volume
	amixer_volume_scontrols = ["Master", "Headset"]
	for sc in amixer_volume_scontrols:
		if sc in scontrols:
			return card, sc
	
	print(f"No suitable amixer volume control for card {card}", file=sys.stderr)
	return card, ""

def get_volume_cmd() -> str:
	card, scontrol = get_current_control()
	return f"amixer -c '{card}' sget '{scontrol}'"

def mute_cmd() -> str:
	card, scontrol = get_current_control()
	return f"amixer -c '{card}' sset '{scontrol}' toggle"

def mic_mute_cmd() -> str:
	card = get_jack_device()
	return f"amixer -c '{card}' sset Capture toggle"

def lower_volume_cmd(step: int) -> str:
	card, scontrol = get_current_control()
	return f"amixer -c '{card}' sset '{scontrol}' {step}-"

def raise_volume_cmd(step: int) -> str:
	card, scontrol = get_current_control()
	return f"amixer -c '{card}' sset '{scontrol}' {step}+"




# jack-alsa-ctl
# Copyright (C) 2022 DCsunset
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import re
from pathlib import Path
import subprocess
import argparse
from ._version import __version__
from .lib import get_volume_cmd, mute_cmd, mic_mute_cmd, raise_volume_cmd, lower_volume_cmd, get_jack_device, list_devices, set_jack_device_cmd

def error(msg: str):
	print(msg, file=sys.stderr)
	sys.exit(1)
	
def run_cmd(cmd: list[str] | str):
	if isinstance(cmd, list):
		for c in cmd:
			subprocess.run(c, shell=True)
	else:
		subprocess.run(cmd, shell=True)


volume_types = ("Capture", "Playback")

def get_volume(volume_type: str):
	if volume_type not in volume_types:
		error(f"Invalid volume type: {volume_type}")

	res = subprocess.run(get_volume_cmd(), shell=True, capture_output=True)
	out = res.stdout.decode("utf-8")
	# filter capture volume
	regex = re.compile(f"{volume_type}.*\\[\\d?\\d?\\d%\\]")
	print("\n".join(
		map(
			lambda l: l.strip(),
			filter(
				lambda l: regex.search(l) is not None,
				out.splitlines()
			)
		)
	))


def main():
	parser = argparse.ArgumentParser(
		description="Control JACK audio with ALSA driver easily via CLI",
		formatter_class=argparse.ArgumentDefaultsHelpFormatter
	)
	parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
	
	# For commands
	sub_parser = parser.add_subparsers(
		dest="command",
		metavar="<command>",
		help="command to execute",
		required=True
	)

	# list_devices
	list_devices_parser = sub_parser.add_parser(
		"list_devices",
		help="List available sound devices"
	)
	list_devices_parser.add_argument(
		"max_num",
		nargs="?",
		default=0,
		type=int,
		help="max number of devices to display (non-positive num means no limit)"
	)
	# get_device
	get_device_parser = sub_parser.add_parser(
		"get_device",
		help="get current device used by JACK server"
	)
	# set device
	set_device_parser = sub_parser.add_parser(
		"set_device",
		help="set device used by JACK server"
	)
	set_device_parser.add_argument(
		"device",
		type=str,
		help="device name"
	)
	# get_volume
	get_volume_parser = sub_parser.add_parser(
		"get_volume",
		help="get current volume"
	)
	get_volume_parser.add_argument(
		"volume_type",
		nargs="?",
		default="Playback",
		choices=["Playback", "Capture"],
		help="get volume of a specific type"
	)
	# mute
	mute_parser = sub_parser.add_parser(
		"mute",
		help="mute current device"
	)
	# mic mute
	mic_mute_parser = sub_parser.add_parser(
		"mic_mute",
		help="mute mic of current device"
	)
	# raise_volume
	raise_volume_parser = sub_parser.add_parser(
		"raise_volume",
		help="raise volume"
	)
	raise_volume_parser.add_argument(
		"step",
		nargs="?",
		default=2,
		type=int,
		help="raise volume by a value"
	)
	# lower_volume
	lower_volume_parser = sub_parser.add_parser(
		"lower_volume",
		help="lower volume"
	)
	lower_volume_parser.add_argument(
		"step",
		nargs="?",
		default=2,
		type=int,
		help="raise volume by a value"
	)

	# Process args
	args = parser.parse_args()
	cmd = args.command
	if cmd == "list_devices":
		print("\n".join(list_devices(args.max_num)))
	elif cmd == "get_device":
		print(get_jack_device())
	elif cmd == "set_device":
		run_cmd(set_jack_device_cmd(args.device))
	elif cmd == "get_volume":
		get_volume(args.volume_type)
	elif cmd == "mute":
		run_cmd(mute_cmd())
	elif cmd == "mic_mute":
		run_cmd(mic_mute_cmd())
	elif cmd == "raise_volume":
		run_cmd(raise_volume_cmd(args.step))
	elif cmd == "lower_volume":
		run_cmd(lower_volume_cmd(args.step))

if __name__ == "__main__":
	main()
